use std::collections::HashMap;
use std::sync::Arc;
use std::time::{Duration, Instant};

use once_cell::sync::Lazy;
use regex::Regex;
use serde::Deserialize;
use tokio::select;
use tokio::sync::broadcast;

use pa_volume_interface::{PaEntityState, PaVolumeInterface};

use crate::model::knob_page::StyleByStateMap;
use crate::model::position::KnobPath;
use crate::modes::knob::{KnobEvent, RotationDirection};
use crate::runner::command::IoWorkerCommand;

#[derive(Debug, Deserialize)]
pub struct Config {
    #[serde(default = "default_delta")]
    pub delta: f32,
    #[serde(with = "serde_regex")]
    pub regex: Regex,
    #[serde(default)]
    pub direction: Direction,
    #[serde(default)]
    pub disable_press_to_mute: bool,
    #[serde(default)]
    pub disable_press_to_unmute: bool,
    #[serde(default)]
    pub muted_turn_action: MutedTurnAction,
    #[serde(default)]
    pub label: HashMap<State, String>,
    #[serde(default)]
    pub style: StyleByStateMap<State>,
}

fn default_delta() -> f32 {
    0.05
}

#[derive(Debug, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum MutedTurnAction {
    #[default]
    Ignore,
    UnmuteAtZero,
    UnmuteAndRestore,
}

#[derive(Debug, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum Direction {
    #[default]
    Output,
    Input,
}

#[derive(Debug, Eq, PartialEq, Hash, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum State {
    Inactive,
    Active,
    Muted,
}

static PA_VOLUME_INTERFACE: Lazy<PaVolumeInterface> = Lazy::new(|| {
    let interface = PaVolumeInterface::spawn_thread(Duration::from_millis(500));
    interface.query_state();
    interface
});

pub async fn handle(path: KnobPath, config: Arc<Config>, mut events: broadcast::Receiver<(KnobPath, KnobEvent)>, commands: flume::Sender<IoWorkerCommand>) {
    let mut volume_update_timestamp = Instant::now();
    let mut value: f32 = 0.0;
    let mut entity_state: Option<Arc<PaEntityState>> = None;

    let pa_volume_interface = &PA_VOLUME_INTERFACE;
    let mut volume_states = pa_volume_interface.subscribe_to_state();

    loop {
        select! {
            Ok(volume_state) = volume_states.recv() => {
                entity_state = volume_state.entities_by_id().values().find(|entity| config.regex.is_match(entity.name())).map(Arc::clone);

                if volume_state.timestamp() > &volume_update_timestamp {
                    value = entity_state.as_ref().map(|s| *s.channel_volumes().first().unwrap()).unwrap_or(0.0);
                    commands.send(IoWorkerCommand::SetKnobValue { path: path.clone(), value }).unwrap();
                }
            }

            Ok((event_path, event)) = events.recv() => {
                if event_path != path {
                    continue;
                }

                if let Some(entity_state) = &entity_state {
                    match event {
                        KnobEvent::Rotate { direction } => {
                            let factor = match direction {
                                RotationDirection::Clockwise => 1.0,
                                RotationDirection::Counterclockwise => -1.0,
                            };

                            volume_update_timestamp = Instant::now();
                            value = (value + (factor * config.delta)).clamp(0.0, 1.0);
                            pa_volume_interface.set_channel_volumes(*entity_state.id(), vec![value; entity_state.channel_volumes().len()]);

                            commands.send(IoWorkerCommand::SetKnobValue { path: path.clone(), value }).unwrap();
                        }
                        KnobEvent::Press => {
                            pa_volume_interface.set_is_muted(*entity_state.id(), !entity_state.is_muted())
                        }
                        _ => {}
                    }
                }
            }
        }
    }
}
